

\noindent Copyright \copyright 1994, 1995, 1996, 1997, 1998 David Neto
\smallskip

\noindent 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
\smallskip

\noindent 
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.
\smallskip

\noindent   
   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the
   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA  02111-1307, USA.
\smallskip

\noindent   
   You may contact David Neto via email at {\tt netod@@acm.org}, or with
   greater latency at
\smallskip
\noindent{\obeylines
     Department of Computer Science
     University of Toronto
     10 King's College Rd.
     Toronto, Ontario
     M5S 3G4
     Canada
}
\medskip


\noindent\hbox{}\hrule\hbox{}\penalty-1000
\vskip0.5cm\relax



@i webdefs.w
@i types.w

{\obeylines
$Log: tabuhash.w,v $
Revision 1.4  1998/08/20 18:33:58  neto
Added debugging output.
Fixed a bug: I wasn't clearing the chain heads in the bounded case.
Unbounded case is untested.

Revision 1.3  1998/08/20 17:01:54  neto
Check for overflow in bounded list if TABUHASH DEBUG is defined.

Revision 1.2  1998/08/20 16:20:30  neto
Fixed to compile.
Then added defenses in the destroy routines.

Revision 1.1  1998/08/20 15:59:08  neto
Initial revision

}

@*Hashed tabu lists.
This module is implements sets of pairs of integers.
These are used by the Lin-Kernighan implementations to keep lists
of forbidden edges.

We'll use the exceedingly simple exclusive-or function as the hash
function.
In symbols, that's |h(u,v)==u^v|. 
We'll resolve collisions with chaining.

Typically, the lists will be small, most times it will have 
fewer than 20 elements.  We'll try to optimize for this behaviour.

A key piece of the puzzle is emptying out the hash table quickly.
We'll keep a |dirty_set_t| object recording which hash slots are used.
When it comes time to empty the table, we'll know where all the 
chains are located.



@ 
Some lists might grow very large indeed.
However, the common case is that the list is guaranteed to remain small 
throughout its lifetime.  In the second case we can preallocate an array
and not have to worry about runtime tests for overflow.  This can 
speed things up.

This module has two implementations to match the two cases.
The routines for the first case have names beginning with 
|tabu_hash_unbd| and routines for the second case 
have names beginning with
|tabu_has_bd|.  However, they have the same interfaces so that the
user can be handed function pointers without regard to which implementation
is being used.  Also, both implementations share the same types.
Some fields will be unused, depending on which implemntation is current
for that object.

@ The outline of this module is as follows:
@c
#include <config.h>
#include "lkconfig.h"
@<System headers@>@;
#include "error.h"
#include "memory.h"
#include "tabuhash.h"

@<Subroutines@>@;
const char *tabuhash_rcs_id = "$Id: tabuhash.w,v 1.4 1998/08/20 18:33:58 neto Exp neto $";

@ We will be using many routines from external libraries.  The interfaces
to those routines are described in the following headers.

@<System headers@>=
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>

@ The exported interface is contained in the \file{tabuhash.h} header file,
which has the following form.

@(tabuhash.h@>=
#if !defined(_TABUHASH_H_)
#define _TABUHASH_H_
extern const char *tabuhash_rcs_id;
@<Module headers required for the interface@>@;
@<Exported type definitions@>@;
@<Exported subroutines@>@;
#endif

@ The interface is as follows:

|tabu_hash_t *tabu_hash_bd_create(int vertex_bound,int max_size)| creates a 
table that can carry up to |max_size| pairs of non-negative integers, where each
integer is less than |vertex_bound|.

|void tabu_hash_bd_destroy(tabu_hash_t *th)| destroys the table.

|int tabu_hash_bd_includes(tabu_hash_t *th, int u, int v)| answers non-zero
if and only if the unordered pair $\{u,v\}$ is present in the list.

|void tabu_hash_bd_add(tabu_hash_t *th, int u, int v)| adds 
the unordered pair $\{u,v\}$ to the list.  
The pair must not already be in the list, or very bad things might happen.

|void tabu_hash_bd_make_empty(tabu_hash_t *th, int u, int v)| removes
all elements from the list.

Procedures
|tabu_hash_unbd_create|,
|tabu_hash_unbd_destroy|,
|tabu_hash_unbd_includes|,
|tabu_hash_unbd_add|,
and
|tabu_hash_unbd_make_empty| do analogous things for unbounded lists.
However, the creation routine ignores its size parameter.

Note that the lookup routines are identical, so
|tabu_hash_unbd_includes| is just a function pointer pointing
to |tabu_hash_bd_includes|.  Clever, no?  (No?!?)
@^function pointer@>

@<Exported subroutines@>=
tabu_hash_t *tabu_hash_bd_create(int vertex_bound, int max_size);
void tabu_hash_bd_destroy(tabu_hash_t *th);
int tabu_hash_bd_includes(tabu_hash_t *th, int u, int v);
void tabu_hash_bd_add(tabu_hash_t *th, int u, int v);
void tabu_hash_bd_make_empty(tabu_hash_t *th);
tabu_hash_t *tabu_hash_unbd_create(int vertex_bound, int max_size);
void tabu_hash_unbd_destroy(tabu_hash_t *th);
int (*tabu_hash_unbd_includes)(tabu_hash_t *th, int u, int v);
void tabu_hash_unbd_add(tabu_hash_t *th, int u, int v);
void tabu_hash_unbd_make_empty(tabu_hash_t *th);

@ The element type is the pair itself, $\{u,v\}$ and a link |next|
pointing to the next element in the hash chain. 
We maintain the the invariant that $u \ge v$.  

The two implementations for the container require different fields.
However, they do share some fields.
Dirty set |dirty| records the set of active chains.
Its domain is zero through |chain_limit-1|.
Array |chain| is the array of chains.

@<Exported type definitions@>=
typedef struct tabu_hash_elem_s {
	int u, v;   /* Invariant: |u >= v|. */
	struct tabu_hash_elem_s *next;
} tabu_hash_elem_t;
@#
typedef struct {
dirty_set_t *dirty;
tabu_hash_elem_t **chain;
int chain_limit;
@<Bounded case fields@>@;
@<Unbounded case fields@>@;
} tabu_hash_t;

@ We need the definition of dirty sets.
@<Module headers required for the interface@>=
#include "dirty.h"

@
The highest index for the |chain| array is 
is one less than the smallest power of two
greater than |vertex_bound-1|.  So |chain_limit| is the smallest power
of two larger than |vertex_bound-1|.

We initialize the |chain| array with zero entries.

We also intiialize the dirty set |dirty| to record which chains are active.
Initially, no chains are active.

@<Common initialization@>=
th = new_of(tabu_hash_t);
th->chain_limit = vertex_bound-1;
th->chain_limit |= th->chain_limit >>1;
th->chain_limit |= th->chain_limit >>2;
th->chain_limit |= th->chain_limit >>4;
th->chain_limit |= th->chain_limit >>8;
th->chain_limit |= th->chain_limit >>16;
#if SIZEOF_INT==8
th->chain_limit |= th->chain_limit >>32;
#endif
errorif(th->chain_limit == INT_MAX, 
	"Too many vertices (%d) to use hashing",vertex_bound);
th->chain_limit += 1;
th->chain = new_arr_of_zero(tabu_hash_elem_t *,th->chain_limit);
th->dirty = dirty_create(th->chain_limit,0,0); /* Empty dirty set. */

@ We need the definition of |INT_MAX|.
@<System headers@>=
#include <limits.h>

@ Cleaning up is the reverse.
@<Common cleanup@>=
free_mem(th->chain);@+mem_deduct(sizeof(tabu_hash_elem_t*)*th->chain_limit);
th->chain_limit=0;
dirty_destroy(th->dirty);
free_mem(th);@+mem_deduct(sizeof(tabu_hash_t));


@ Searching for an element is the same for both implementations.
We follow examine the chain of pairs we find at the hash table 
address |u ^v|.

We take advantage of the invariant within the data structure, \ie
|u >= v|, to reduce the number of operations within the loop.

@<Subroutines@>=
int 
tabu_hash_bd_includes(tabu_hash_t *th, int u, int v) 
{
	tabu_hash_elem_t *here;
	int uu = u, vv=v;
	if (uu < vv) { @+ uu=v; @+ vv=u; @+ } /* Satisfy invariant. */
	for (here=th->chain[u^v];here;here=here->next) {
		@<Verbose output: checking |here| for |u,v|@>@;
		if ( here->u == uu && here->v == vv ) {
			@<Verbose output: tabu list includes |u,v|@>@;
			return 1;
		}
	}
	@<Verbose output: tabu list doesn't include |u,v|@>@;
	return 0;  /* Not found. */
}

int (*tabu_hash_unbd_includes)(tabu_hash_t *th, int u, int v) 
	= tabu_hash_bd_includes;

@*Bounded lists.
In the bounded case, the elements are stored in a pre-allocated array.
Since we can only add individuals or wipe out the array entirely, we
can just accumulate them at the beginning of the array.

Field |thread| is the array holding all the threaded chains.

Field |size| contains the number of elements already in the array.

Field |max_size| records the maximum size of the array.

@<Bounded case fields@>=
tabu_hash_elem_t *thread;
int size;
int max_size;

@ The creation routine includes the common initializations we gave above.
@<Subroutines@>=
tabu_hash_t *
tabu_hash_bd_create(int vertex_bound,int max_size)
{
	tabu_hash_t *th;
	@<Common initialization@>@;
	errorif(max_size <= 0, "Maximum tabu list size (%d) must be positive.",max_size);
	th->thread = new_arr_of(tabu_hash_elem_t,max_size);
	th->max_size = max_size;
	th->size=0;
	return th;
}

@ The destruction routine is likewise.
@<Subroutines@>=
void
tabu_hash_bd_destroy(tabu_hash_t *th)
{
	if ( th ) {
		free_mem(th->thread);@+mem_deduct(sizeof(tabu_hash_elem_t *)*th->max_size);
		th->max_size = 0;
		th->size=0;
		@<Common cleanup@>@;
	}
}

@ Adding in the bounded case is then easy.
We just fill in the next slot in the stack-like |thread| array, and
thread it with the previous slot this hashed chain used to point to.

Of course, we mark the chain dirty if it wasn't already.
We perform the test with |was_dirty| in order to save a procedure call
in many cases.

@<Subroutines@>=
void 
tabu_hash_bd_add(tabu_hash_t *th, int u, int v)
{
	const int h = u^v;
	tabu_hash_elem_t *was_dirty;
	int uu=u, vv=v;
#if defined(TABUHASH_DEBUG)
	errorif(th->size >= th->max_size, 
		"Bounded tabu list already saturated with %d elements.",th->size);
#endif
	if  ( u < v ) { @+ uu = v; @+ vv=u; @+ } /* Satisfy invariant. */
	th->thread[th->size].u = uu;
	th->thread[th->size].v = vv;
	was_dirty = th->thread[th->size].next = th->chain[h];
	th->chain[h] = th->thread + th->size;
	th->size++;
	if ( !was_dirty ) dirty_add(th->dirty,h);	/* Mark this thread dirty. */
	@<Verbose: announce bounded add@>@;
}

@ Making it empty is just a matter of resetting the high water mark
(|th->size|), resetting the heads of the chains to zero, and
the emptying out the dirty set.  We fully consume the dirty set when
we reset the heads of chains.

@<Subroutines@>=
void 
tabu_hash_bd_make_empty(tabu_hash_t *th)
{
	int c;
	dirty_set_t *dirty = th->dirty;
	while ( (c = dirty_remove(dirty)) >= 0 ) th->chain[c]=NULL;
	th->size=0;
	@<Verbose: tabu bounded make empty@>@;
}


@*Unbounded lists.
In the unbounded case we use a pool of elements instead of a preallocated
array.

Field |elem_pool| is the pool holding all the elements.

@<Unbounded case fields@>=
pool_t *elem_pool;

@
@<Module headers required for the interface@>=
#include "pool.h"

@ The creation routine again includes the common initializations we gave above.

If the |max_size| parameter is positive, then we take it
as a hint for the number of elements
we should preallocate in the pool.  The pool will expand as necessary.
If |max_size| is zero or less, then we pre-allocate 120 elements.

@<Subroutines@>=
tabu_hash_t *
tabu_hash_unbd_create(int vertex_bound,int max_size)
{
	tabu_hash_t *th;
	@<Common initialization@>@;
	th->elem_pool=
		pool_create(sizeof(tabu_hash_elem_t),max_size > 0 ? max_size : 120 );
	return th;
}

@ The destruction routine is likewise.
@<Subroutines@>=
void
tabu_hash_unbd_destroy(tabu_hash_t *th)
{
	if ( th ) {
		pool_destroy(th->elem_pool);
		@<Common cleanup@>@;
	}
}

@ Adding in the unbounded is similar to the bounded case.
Except we get the fresh element from the pool instead of from the 
|thread| array.
@<Subroutines@>=
void 
tabu_hash_unbd_add(tabu_hash_t *th, int u, int v)
{
	const int h = u^v;
	tabu_hash_elem_t *e = pool_alloc(th->elem_pool);
	tabu_hash_elem_t *was_dirty;
	int uu=u, vv=v;
	if  ( u < v ) { @+ uu = v; @+ vv=u; @+ }
	e->u = uu;
	e->v = vv;
	was_dirty = e->next = th->chain[h];
	th->chain[h] = e;
	if ( !was_dirty ) dirty_add(th->dirty,h);	/* Mark this thread dirty. */
	@<Verbose: announce unbounded add@>@;
}

@ Making it empty is just emptying the pool and the dirty set.

In the common case there will be few dirty chains and each one of them
will be short.  So instead of traversing the entire chain array we just
enumerate them through the |dirty| set.  Furthermore, for each chain
we just follow the (hopefully short) chain of pointers, deallocating
pool elements along the way.  

This is a bit more complex than just blasting away at the pool and the
array, but it touches only the data that really needs to be touched.

@<Subroutines@>=
void 
tabu_hash_unbd_make_empty(tabu_hash_t *th)
{
	int c;
	pool_t *pool=th->elem_pool;
	dirty_set_t *dirty = th->dirty;
	while ( (c = dirty_remove(dirty)) >= 0 ) {
		tabu_hash_elem_t *here;
		for ( here=th->chain[c]; here ; here=here->next ) pool_free(pool,here);
		th->chain[c]=NULL;
	}
	@<Verbose: tabu unbounded make empty@>@;
}

@*Verbose output.
@<Verbose: announce bounded add@>=
#if TABUHASH_MAX_VERBOSE >= 160
{ extern int verbose;
if (verbose >= 160 ) { printf("tabu_hash_bd_add: Added %d %d\n",u,v); }
if (verbose >= 500 ) {
	@<Verbose: print the bounded tabu list@>@;
}
}
#endif

@
@<Verbose: announce unbounded add@>=
#if TABUHASH_MAX_VERBOSE >= 160
{ extern int verbose;
if (verbose >= 160 ) { printf("tabu_hash_unbd_add: Added %d %d\n",u,v); }
if (verbose >= 500 ) {
	@<Verbose: print the unbounded tabu list@>@;
}
}
#endif


@  How do I do this well?
@<Verbose: print the bounded tabu list@>=

@
@<Verbose: print the unbounded tabu list@>=

@
@<Verbose output: tabu list doesn't include |u,v|@>=
#if TABUHASH_MAX_VERBOSE >= 180
{ extern int verbose;
if (verbose >= 180 ) { printf("tabu_hash_includes: %d %d is not tabu\n",u,v); }
}
#endif


@
@<Verbose output: tabu list includes |u,v|@>=
#if TABUHASH_MAX_VERBOSE >= 180
{ extern int verbose;
if (verbose >= 180 ) { printf("tabu_hash_includes: %d %d is tabu\n",u,v); }
}
#endif

@
@<Verbose: tabu bounded make empty@>=
#if TABUHASH_MAX_VERBOSE >= 160
{ extern int verbose;
if (verbose >= 160 ) { printf("tabu_hash_bd_make_empty\n"); }
}
#endif

@
@<Verbose: tabu unbounded make empty@>=
#if TABUHASH_MAX_VERBOSE >= 160
{ extern int verbose;
if (verbose >= 160 ) { printf("tabu_hash_unbd_make_empty\n"); }
}
#endif

@
@<Verbose output: checking |here| for |u,v|@>=
#if TABUHASH_MAX_VERBOSE >= 160
{ extern int verbose;
if (verbose >= 200 ) { 
	printf("tabu_hash_includes: check chain %p for %d %d\n",here,uu,vv); 
}
}
#endif

@*Index.
